@@ -5,6 +5,7 @@ from peewee import Database
 from peewee import ModelIndex
 
 from .base import get_in_memory_db
+from .base import requires_pglike
 from .base import BaseTestCase
 from .base import ModelDatabaseTestCase
 from .base import TestModel
@@ -12,9 +13,17 @@ from .base import __sql__
 from .base_models import *
 
 
+class CKM(TestModel):
+    category = CharField()
+    key = CharField()
+    value = IntegerField()
+    class Meta:
+        primary_key = CompositeKey('category', 'key')
+
+
 class TestModelSQL(ModelDatabaseTestCase):
     database = get_in_memory_db()
-    requires = [Category, Note, Person, Relationship]
+    requires = [Category, CKM, Note, Person, Relationship, Sample, User]
 
     def test_select(self):
         query = (Person
@@ -78,6 +87,48 @@ class TestModelSQL(ModelDatabaseTestCase):
             'LEFT OUTER JOIN "tweet" AS "t2" ON ("t2"."user_id" = "t1"."id") '
             'GROUP BY "t1"."id", "t1"."username"'), [])
 
+    def test_group_by_extend(self):
+        query = (User
+                 .select(User, fn.COUNT(Tweet.id).alias('tweet_count'))
+                 .join(Tweet, JOIN.LEFT_OUTER)
+                 .group_by_extend(User.id).group_by_extend(User.username))
+        self.assertSQL(query, (
+            'SELECT "t1"."id", "t1"."username", '
+            'COUNT("t2"."id") AS "tweet_count" '
+            'FROM "users" AS "t1" '
+            'LEFT OUTER JOIN "tweet" AS "t2" ON ("t2"."user_id" = "t1"."id") '
+            'GROUP BY "t1"."id", "t1"."username"'), [])
+
+    def test_order_by(self):
+        query = (User
+                 .select()
+                 .order_by(User.username.desc(), User.id))
+        self.assertSQL(query, (
+            'SELECT "t1"."id", "t1"."username" FROM "users" AS "t1" '
+            'ORDER BY "t1"."username" DESC, "t1"."id"'), [])
+
+    def test_order_by_extend(self):
+        query = (User
+                 .select()
+                 .order_by_extend(User.username.desc())
+                 .order_by_extend(User.id))
+        self.assertSQL(query, (
+            'SELECT "t1"."id", "t1"."username" FROM "users" AS "t1" '
+            'ORDER BY "t1"."username" DESC, "t1"."id"'), [])
+
+    def test_paginate(self):
+        # Get the first page, default is limit of 20.
+        query = User.select().paginate(1)
+        self.assertSQL(query, (
+            'SELECT "t1"."id", "t1"."username" FROM "users" AS "t1" '
+            'LIMIT ? OFFSET ?'), [20, 0])
+
+        # Page 3 contains rows 31-45.
+        query = User.select().paginate(3, 15)
+        self.assertSQL(query, (
+            'SELECT "t1"."id", "t1"."username" FROM "users" AS "t1" '
+            'LIMIT ? OFFSET ?'), [15, 30])
+
     def test_subquery_correction(self):
         users = User.select().where(User.username.in_(['foo', 'bar']))
         query = Tweet.select().where(Tweet.user.in_(users))
@@ -222,6 +273,89 @@ class TestModelSQL(ModelDatabaseTestCase):
             'SELECT "t1"."name", "t1"."parent_id" FROM "category" AS "t1" '
             'WHERE ("t1"."parent_id" = ?)'), ['test'])
 
+    def test_cross_join(self):
+        class A(TestModel):
+            id = IntegerField(primary_key=True)
+        class B(TestModel):
+            id = IntegerField(primary_key=True)
+        query = (A
+                 .select(A.id.alias('aid'), B.id.alias('bid'))
+                 .join(B, JOIN.CROSS)
+                 .order_by(A.id, B.id))
+        self.assertSQL(query, (
+            'SELECT "t1"."id" AS "aid", "t2"."id" AS "bid" '
+            'FROM "a" AS "t1" '
+            'CROSS JOIN "b" AS "t2" '
+            'ORDER BY "t1"."id", "t2"."id"'), [])
+
+    def test_join_expr(self):
+        class User(TestModel):
+            username = TextField(primary_key=True)
+        class Tweet(TestModel):
+            user = ForeignKeyField(User, backref='tweets')
+            content = TextField()
+
+        sql = ('SELECT "t1"."id", "t1"."user_id", "t1"."content", '
+               '"t2"."username" FROM "tweet" AS "t1" '
+               'INNER JOIN "user" AS "t2" '
+               'ON ("t1"."user_id" = "t2"."username")')
+
+        query = Tweet.select(Tweet, User).join(User)
+        self.assertSQL(query, sql, [])
+
+        query = (Tweet
+                 .select(Tweet, User)
+                 .join(User, on=(Tweet.user == User.username)))
+        self.assertSQL(query, sql, [])
+
+        join_expr = ((Tweet.user == User.username) & (Value(1) == 1))
+        query = Tweet.select().join(User, on=join_expr)
+        self.assertSQL(query, (
+            'SELECT "t1"."id", "t1"."user_id", "t1"."content" '
+            'FROM "tweet" AS "t1" '
+            'INNER JOIN "user" AS "t2" '
+            'ON (("t1"."user_id" = "t2"."username") AND (? = ?))'), [1, 1])
+
+    def test_join_multiple_fks(self):
+        class A(TestModel):
+            name = TextField()
+        class B(TestModel):
+            name = TextField(primary_key=True)
+            a1 = ForeignKeyField(A, backref='b_set1')
+            a2 = ForeignKeyField(A, field=A.name, backref='b_set2')
+
+        A1 = A.alias('a1')
+        A2 = A.alias('a2')
+
+        sql = ('SELECT "t1"."name", "t1"."a1_id", "t1"."a2_id", '
+               '"a1"."id", "a1"."name", "a2"."id", "a2"."name" '
+               'FROM "b" AS "t1" '
+               'INNER JOIN "a" AS "a1" ON ("t1"."a1_id" = "a1"."id") '
+               'INNER JOIN "a" AS "a2" ON ("t1"."a2_id" = "a2"."name")')
+
+        query = (B.select(B, A1, A2)
+                 .join_from(B, A1, on=B.a1)
+                 .join_from(B, A2, on=B.a2))
+        self.assertSQL(query, sql, [])
+
+        query = (B.select(B, A1, A2)
+                 .join(A1, on=(B.a1 == A1.id)).switch(B)
+                 .join(A2, on=(B.a2 == A2.name)))
+        self.assertSQL(query, sql, [])
+
+        jx1 = (B.a1 == A1.id) & (Value(1) == 1)
+        jx2 = (Value(1) == 1) & (B.a2 == A2.name)
+        query = (B.select()
+                 .join(A1, on=jx1).switch(B)
+                 .join(A2, on=jx2))
+        self.assertSQL(query, (
+            'SELECT "t1"."name", "t1"."a1_id", "t1"."a2_id" '
+            'FROM "b" AS "t1" '
+            'INNER JOIN "a" AS "a1" '
+            'ON (("t1"."a1_id" = "a1"."id") AND (? = ?)) '
+            'INNER JOIN "a" AS "a2" '
+            'ON ((? = ?) AND ("t1"."a2_id" = "a2"."name"))'), [1, 1, 1, 1])
+
     def test_raw(self):
         query = (Person
                  .raw('SELECT first, last, dob FROM person '
@@ -281,6 +415,47 @@ class TestModelSQL(ModelDatabaseTestCase):
             'VALUES (?, ?), (?, ?)'),
             [1, 'note-1', 2, 'note-2'])
 
+    def test_insert_many_defaults(self):
+        # Verify fields are inferred and values are read correctly, when
+        # partial data is given and a field has default values.
+        s2 = {'counter': 2, 'value': 2.}
+        s3 = {'counter': 3}
+        self.assertSQL(Sample.insert_many([s2, s3]), (
+            'INSERT INTO "sample" ("counter", "value") VALUES (?, ?), (?, ?)'),
+            [2, 2., 3, 1.])
+
+        self.assertSQL(Sample.insert_many([s3, s2]), (
+            'INSERT INTO "sample" ("counter", "value") VALUES (?, ?), (?, ?)'),
+            [3, 1., 2, 2.])
+
+    def test_insert_many_list_with_fields(self):
+        data = [(i,) for i in ('charlie', 'huey', 'zaizee')]
+        query = User.insert_many(data, fields=[User.username])
+        self.assertSQL(query, (
+            'INSERT INTO "users" ("username") VALUES (?), (?), (?)'),
+            ['charlie', 'huey', 'zaizee'])
+
+        # Use field name instead of field obj.
+        query = User.insert_many(data, fields=['username'])
+        self.assertSQL(query, (
+            'INSERT INTO "users" ("username") VALUES (?), (?), (?)'),
+            ['charlie', 'huey', 'zaizee'])
+
+    def test_insert_many_infer_fields(self):
+        data = [('f1', 'l1', '1980-01-01'),
+                ('f2', 'l2', '1980-02-02')]
+        self.assertSQL(Person.insert_many(data), (
+            'INSERT INTO "person" ("first", "last", "dob") '
+            'VALUES (?, ?, ?), (?, ?, ?)'),
+            ['f1', 'l1', datetime.date(1980, 1, 1),
+             'f2', 'l2', datetime.date(1980, 2, 2)])
+
+        # When primary key is not auto-increment, PKs are included.
+        data = [('c1', 'k1', 1), ('c2', 'k2', 2)]
+        self.assertSQL(CKM.insert_many(data), (
+            'INSERT INTO "ckm" ("category", "key", "value") '
+            'VALUES (?, ?, ?), (?, ?, ?)'), ['c1', 'k1', 1, 'c2', 'k2', 2])
+
     def test_insert_query(self):
         select = (Person
                   .select(Person.id, Person.first)
@@ -309,7 +484,7 @@ class TestModelSQL(ModelDatabaseTestCase):
         query = User.insert({User.username: 'zaizee'})
         self.assertSQL(query, (
             'INSERT INTO "user" ("username") '
-            'VALUES (?) RETURNING "id"'), ['zaizee'])
+            'VALUES (?) RETURNING "user"."id"'), ['zaizee'])
 
         class Person(Model):
             name = CharField()
@@ -320,12 +495,148 @@ class TestModelSQL(ModelDatabaseTestCase):
         query = Person.insert({Person.name: 'charlie', Person.ssn: '123'})
         self.assertSQL(query, (
             'INSERT INTO "person" ("ssn", "name") VALUES (?, ?) '
-            'RETURNING "ssn"'), ['123', 'charlie'])
+            'RETURNING "person"."ssn"'), ['123', 'charlie'])
 
         query = Person.insert({Person.name: 'huey'}).returning()
         self.assertSQL(query, (
             'INSERT INTO "person" ("name") VALUES (?)'), ['huey'])
 
+        query = (Person
+                 .insert({Person.name: 'foo'})
+                 .returning(Person.ssn.alias('new_ssn')))
+        self.assertSQL(query, (
+            'INSERT INTO "person" ("name") VALUES (?) '
+            'RETURNING "person"."ssn" AS "new_ssn"'), ['foo'])
+
+    def test_insert_get_field_values(self):
+        class User(TestModel):
+            username = TextField(primary_key=True)
+            class Meta:
+                database = self.database
+
+        class Tweet(TestModel):
+            user = ForeignKeyField(User)
+            content = TextField()
+            class Meta:
+                database = self.database
+
+        queries = (
+            User.insert(username='a'),
+            User.insert({'username': 'a'}),
+            User.insert({User.username: 'a'}))
+        for query in queries:
+            self.assertSQL(query, ('INSERT INTO "user" ("username") '
+                                   'VALUES (?)'), ['a'])
+
+        # Verify that we can provide all kinds of combinations to the
+        # constructor to INSERT and it will map the parameters correctly
+        # without losing values.
+        a = User(username='a')
+        queries = (
+            Tweet.insert(user=a, content='ca'),
+            Tweet.insert({'user': a, 'content': 'ca'}),
+            Tweet.insert({Tweet.user: a, 'content': 'ca'}),
+            Tweet.insert({'user': a, Tweet.content: 'ca'}),
+            Tweet.insert({Tweet.user: a, Tweet.content: 'ca'}),
+            Tweet.insert({Tweet.user: a}, content='ca'),
+            Tweet.insert({Tweet.content: 'ca'}, user=a),
+            Tweet.insert({'user': a}, content='ca'),
+            Tweet.insert({'content': 'ca'}, user=a),
+
+            # Also test using the foreign-key descriptor and column name.
+            Tweet.insert({Tweet.user_id: a, Tweet.content: 'ca'}),
+            Tweet.insert(user_id=a, content='ca'),
+            Tweet.insert({'user_id': a, 'content': 'ca'}))
+
+        for query in queries:
+            self.assertSQL(query, ('INSERT INTO "tweet" ("user_id", "content")'
+                                   ' VALUES (?, ?)'), ['a', 'ca'])
+
+    def test_insert_many_get_field_values(self):
+        class User(TestModel):
+            username = TextField(primary_key=True)
+            class Meta:
+                database = self.database
+
+        class Tweet(TestModel):
+            user = ForeignKeyField(User)
+            content = TextField()
+            class Meta:
+                database = self.database
+
+        # Ensure we can handle any combination of insert-data key and field
+        # list value.
+        pairs = ((User.username, 'username'),
+                 ('username', User.username),
+                 ('username', 'username'),
+                 (User.username, User.username))
+
+        for dict_key, fields_key in pairs:
+            iq = User.insert_many([{dict_key: u} for u in 'abc'],
+                                  fields=[fields_key])
+            self.assertSQL(iq, (
+                'INSERT INTO "user" ("username") VALUES (?), (?), (?)'),
+                ['a', 'b', 'c'])
+
+        a, b = User(username='a'), User(username='b')
+        user_content = (
+            (a, 'ca1'),
+            (a, 'ca2'),
+            (b, 'cb1'),
+            ('a', 'ca3'))  # Specify user id directly.
+
+        # Ensure we can mix-and-match key type within insert-data.
+        pairs = (('user', 'content'),
+                 (Tweet.user, Tweet.content),
+                 (Tweet.user, 'content'),
+                 ('user', Tweet.content),
+                 ('user_id', 'content'),
+                 (Tweet.user_id, Tweet.content))
+
+        for ukey, ckey in pairs:
+            iq = Tweet.insert_many([{ukey: u, ckey: c}
+                                    for u, c in user_content])
+            self.assertSQL(iq, (
+                'INSERT INTO "tweet" ("user_id", "content") VALUES '
+                '(?, ?), (?, ?), (?, ?), (?, ?)'),
+                ['a', 'ca1', 'a', 'ca2', 'b', 'cb1', 'a', 'ca3'])
+
+    def test_insert_many_dict_and_list(self):
+        class R(TestModel):
+            k = TextField(column_name='key')
+            v = IntegerField(column_name='value', default=0)
+            class Meta:
+                database = self.database
+
+        data = (
+            {'k': 'k1', 'v': 1},
+            {R.k: 'k2', R.v: 2},
+            {'key': 'k3', 'value': 3},
+            ('k4', 4),
+            ('k5', '5'),  # Will be converted properly.
+            {R.k: 'k6', R.v: '6'},
+            {'key': 'k7', 'value': '7'},
+            {'k': 'kx'},
+            ('ky',))
+
+        param_str = ', '.join('(?, ?)' for _ in range(len(data)))
+        queries = (
+            R.insert_many(data),
+            R.insert_many(data, fields=[R.k, R.v]),
+            R.insert_many(data, fields=['k', 'v']))
+        for query in queries:
+            self.assertSQL(query, (
+                'INSERT INTO "r" ("key", "value") VALUES %s' % param_str),
+                ['k1', 1, 'k2', 2, 'k3', 3, 'k4', 4, 'k5', 5, 'k6', 6,
+                 'k7', 7, 'kx', 0, 'ky', 0])
+
+    def test_insert_modelalias(self):
+        UA = User.alias('ua')
+        self.assertSQL(UA.insert({UA.username: 'huey'}), (
+            'INSERT INTO "users" ("username") VALUES (?)'), ['huey'])
+        self.assertSQL(UA.insert(username='huey'), (
+            'INSERT INTO "users" ("username") VALUES (?)'), ['huey'])
+
     def test_update(self):
         class Stat(TestModel):
             url = TextField()
@@ -337,18 +648,44 @@ class TestModelSQL(ModelDatabaseTestCase):
                           Stat.timestamp: datetime.datetime(2017, 1, 1)})
                  .where(Stat.url == '/peewee'))
         self.assertSQL(query, (
-            'UPDATE "stat" SET "count" = ("count" + ?), "timestamp" = ? '
-            'WHERE ("url" = ?)'),
+            'UPDATE "stat" SET "count" = ("stat"."count" + ?), '
+            '"timestamp" = ? '
+            'WHERE ("stat"."url" = ?)'),
             [1, 1483228800, '/peewee'])
 
         query = (Stat
                  .update(count=Stat.count + 1)
                  .where(Stat.url == '/peewee'))
         self.assertSQL(query, (
-            'UPDATE "stat" SET "count" = ("count" + ?) '
-            'WHERE ("url" = ?)'),
+            'UPDATE "stat" SET "count" = ("stat"."count" + ?) '
+            'WHERE ("stat"."url" = ?)'),
             [1, '/peewee'])
 
+    def test_update_subquery(self):
+        class U(TestModel):
+            username = TextField()
+            flood_count = IntegerField()
+
+        class T(TestModel):
+            user = ForeignKeyField(U)
+
+        ctq = T.select(fn.COUNT(T.id) / 100).where(T.user == U.id)
+        subq = (T
+                .select(T.user)
+                .group_by(T.user)
+                .having(fn.COUNT(T.id) > 100))
+        query = (U
+                 .update({U.flood_count: ctq})
+                 .where(U.id.in_(subq)))
+        self.assertSQL(query, (
+            'UPDATE "u" SET "flood_count" = ('
+            'SELECT (COUNT("t1"."id") / ?) FROM "t" AS "t1" '
+            'WHERE ("t1"."user_id" = "u"."id")) '
+            'WHERE ("u"."id" IN ('
+            'SELECT "t1"."user_id" FROM "t" AS "t1" '
+            'GROUP BY "t1"."user_id" '
+            'HAVING (COUNT("t1"."id") > ?)))'), [100, 100])
+
     def test_update_from(self):
         class SalesPerson(TestModel):
             first = TextField()
@@ -366,26 +703,26 @@ class TestModelSQL(ModelDatabaseTestCase):
                  .where(Account.sales == SalesPerson.id))
         self.assertSQL(query, (
             'UPDATE "account" SET '
-            '"contact_first" = "first", '
-            '"contact_last" = "last" '
+            '"contact_first" = "t1"."first", '
+            '"contact_last" = "t1"."last" '
             'FROM "sales_person" AS "t1" '
-            'WHERE ("sales_id" = "id")'), [])
+            'WHERE ("account"."sales_id" = "t1"."id")'), [])
 
         query = (User
-                 .update({User.username: QualifiedNames(Tweet.content)})
+                 .update({User.username: Tweet.content})
                  .from_(Tweet)
                  .where(Tweet.content == 'tx'))
         self.assertSQL(query, (
             'UPDATE "users" SET "username" = "t1"."content" '
-            'FROM "tweet" AS "t1" WHERE ("content" = ?)'), ['tx'])
+            'FROM "tweet" AS "t1" WHERE ("t1"."content" = ?)'), ['tx'])
 
     def test_update_from_qualnames(self):
         data = [(1, 'u1-x'), (2, 'u2-x')]
         vl = ValuesList(data, columns=('id', 'username'), alias='tmp')
         query = (User
-                 .update({User.username: QualifiedNames(vl.c.username)})
+                 .update({User.username: vl.c.username})
                  .from_(vl)
-                 .where(QualifiedNames(User.id == vl.c.id)))
+                 .where(User.id == vl.c.id))
         self.assertSQL(query, (
             'UPDATE "users" SET "username" = "tmp"."username" '
             'FROM (VALUES (?, ?), (?, ?)) AS "tmp"("id", "username") '
@@ -396,9 +733,9 @@ class TestModelSQL(ModelDatabaseTestCase):
         vl = ValuesList(data, columns=('id', 'username'), alias='tmp')
         subq = vl.select(vl.c.id, vl.c.username)
         query = (User
-                 .update({User.username: QualifiedNames(subq.c.username)})
+                 .update({User.username: subq.c.username})
                  .from_(subq)
-                 .where(QualifiedNames(User.id == subq.c.id)))
+                 .where(User.id == subq.c.id))
         self.assertSQL(query, (
             'UPDATE "users" SET "username" = "t1"."username" FROM ('
             'SELECT "tmp"."id", "tmp"."username" '
@@ -411,13 +748,13 @@ class TestModelSQL(ModelDatabaseTestCase):
                  .where(Note.author << (Person.select(Person.id)
                                         .where(Person.last == 'cat'))))
         self.assertSQL(query, ('DELETE FROM "note" '
-                               'WHERE ("author_id" IN ('
+                               'WHERE ("note"."author_id" IN ('
                                'SELECT "t1"."id" FROM "person" AS "t1" '
                                'WHERE ("t1"."last" = ?)))'), ['cat'])
 
         query = Note.delete().where(Note.author == Person(id=123))
-        self.assertSQL(query, 'DELETE FROM "note" WHERE ("author_id" = ?)',
-                       [123])
+        self.assertSQL(query, (
+            'DELETE FROM "note" WHERE ("note"."author_id" = ?)'), [123])
 
     def test_delete_recursive(self):
         class User(TestModel):
@@ -440,20 +777,20 @@ class TestModelSQL(ModelDatabaseTestCase):
 
         self.assertEqual(sorted(accum), [
             ('DELETE FROM "like" WHERE ('
-             '"tweet_id" IN ('
+             '"like"."tweet_id" IN ('
              'SELECT "t1"."id" FROM "tweet" AS "t1" WHERE ('
              '"t1"."user_id" = ?)))', [1]),
-            ('DELETE FROM "like" WHERE ("user_id" = ?)', [1]),
-            ('DELETE FROM "relationship" WHERE ("from_user_id" = ?)', [1]),
-            ('DELETE FROM "relationship" WHERE ("to_user_id" = ?)', [1]),
-            ('DELETE FROM "tweet" WHERE ("user_id" = ?)', [1]),
+            ('DELETE FROM "like" WHERE ("like"."user_id" = ?)', [1]),
+            ('DELETE FROM "relationship" '
+             'WHERE ("relationship"."from_user_id" = ?)', [1]),
+            ('DELETE FROM "relationship" '
+             'WHERE ("relationship"."to_user_id" = ?)', [1]),
+            ('DELETE FROM "tweet" WHERE ("tweet"."user_id" = ?)', [1]),
         ])
 
     def test_aliases(self):
         class A(TestModel):
             a = CharField()
-            class Meta:
-                table_alias = 'a_tbl'
         class B(TestModel):
             b = CharField()
             a_link = ForeignKeyField(A)
@@ -463,8 +800,6 @@ class TestModelSQL(ModelDatabaseTestCase):
         class D(TestModel):
             d = CharField()
             c_link = ForeignKeyField(C)
-            class Meta:
-                table_alias = 'd_tbl'
 
         query = (D
                  .select(D.d, C.c)
@@ -472,13 +807,13 @@ class TestModelSQL(ModelDatabaseTestCase):
                  .where(C.b_link << (
                      B.select(B.id).join(A).where(A.a == 'a'))))
         self.assertSQL(query, (
-            'SELECT "d_tbl"."d", "t1"."c" '
-            'FROM "d" AS "d_tbl" '
-            'INNER JOIN "c" AS "t1" ON ("d_tbl"."c_link_id" = "t1"."id") '
-            'WHERE ("t1"."b_link_id" IN ('
-            'SELECT "t2"."id" FROM "b" AS "t2" '
-            'INNER JOIN "a" AS "a_tbl" ON ("t2"."a_link_id" = "a_tbl"."id") '
-            'WHERE ("a_tbl"."a" = ?)))'), ['a'])
+            'SELECT "t1"."d", "t2"."c" '
+            'FROM "d" AS "t1" '
+            'INNER JOIN "c" AS "t2" ON ("t1"."c_link_id" = "t2"."id") '
+            'WHERE ("t2"."b_link_id" IN ('
+            'SELECT "t3"."id" FROM "b" AS "t3" '
+            'INNER JOIN "a" AS "t4" ON ("t3"."a_link_id" = "t4"."id") '
+            'WHERE ("t4"."a" = ?)))'), ['a'])
 
     def test_schema(self):
         class WithSchema(TestModel):
@@ -493,6 +828,52 @@ class TestModelSQL(ModelDatabaseTestCase):
             'WHERE ("t1"."data" = ?)'), ['zaizee'])
 
 
+@requires_pglike
+class TestOnConflictSQL(ModelDatabaseTestCase):
+    requires = [Emp, OCTest, UKVP]
+
+    def test_atomic_update(self):
+        query = OCTest.insert(a='foo', b=1).on_conflict(
+            conflict_target=(OCTest.a,),
+            update={OCTest.b: OCTest.b + 2})
+
+        self.assertSQL(query, (
+            'INSERT INTO "oc_test" ("a", "b", "c") VALUES (?, ?, ?) '
+            'ON CONFLICT ("a") '
+            'DO UPDATE SET "b" = ("oc_test"."b" + ?) '
+            'RETURNING "oc_test"."id"'), ['foo', 1, 0, 2])
+
+    def test_update_where_clause(self):
+        # Add a new row with the given "a" value. If a conflict occurs,
+        # re-insert with b=b+2 so long as the original b < 3.
+        query = OCTest.insert(a='foo', b=1).on_conflict(
+            conflict_target=(OCTest.a,),
+            update={OCTest.b: OCTest.b + 2},
+            where=(OCTest.b < 3))
+        self.assertSQL(query, (
+            'INSERT INTO "oc_test" ("a", "b", "c") VALUES (?, ?, ?) '
+            'ON CONFLICT ("a") DO UPDATE SET "b" = ("oc_test"."b" + ?) '
+            'WHERE ("oc_test"."b" < ?) '
+            'RETURNING "oc_test"."id"'), ['foo', 1, 0, 2, 3])
+
+    def test_conflict_target_constraint_where(self):
+        fields = [UKVP.key, UKVP.value, UKVP.extra]
+        data = [('k1', 1, 2), ('k2', 2, 3)]
+
+        query = (UKVP.insert_many(data, fields)
+                 .on_conflict(conflict_target=(UKVP.key, UKVP.value),
+                              conflict_where=(UKVP.extra > 1),
+                              preserve=(UKVP.extra,),
+                              where=(UKVP.key != 'kx')))
+        self.assertSQL(query, (
+            'INSERT INTO "ukvp" ("key", "value", "extra") '
+            'VALUES (?, ?, ?), (?, ?, ?) '
+            'ON CONFLICT ("key", "value") WHERE ("extra" > ?) '
+            'DO UPDATE SET "extra" = EXCLUDED."extra" '
+            'WHERE ("ukvp"."key" != ?) RETURNING "ukvp"."id"'),
+            ['k1', 1, 2, 'k2', 2, 3, 1, 'kx'])
+
+
 class TestStringsForFieldsa(ModelDatabaseTestCase):
     database = get_in_memory_db()
     requires = [Note, Person, Relationship]
@@ -520,7 +901,7 @@ class TestStringsForFieldsa(ModelDatabaseTestCase):
         qliteral = Person.update({'last': 'kitty'}).where(Person.last == 'cat')
         for query in (qkwargs, qliteral):
             self.assertSQL(query, (
-                'UPDATE "person" SET "last" = ? WHERE ("last" = ?)'),
+                'UPDATE "person" SET "last" = ? WHERE ("person"."last" = ?)'),
                 ['kitty', 'cat'])
 
 
